Module browser

Module browser 

Re-exports

pub use tab::Tab;
pub use transport::ConnectionClosed;
pub use B::GetVersionReturnObject;

Modules

context
tab
transport

Structs

Browser
A handle to an instance of Chrome / Chromium, which wraps a WebSocket connection to its debugging port.
BrowserInner
LaunchOptions
Represents the way in which Chrome is run. By default it will search for a Chrome binary on the system, use an available port for debugging, and start in headless mode.
LaunchOptionsBuilder
Builder for LaunchOptions.

Statics

DEFAULT_ARGS
These are passed to the Chrome binary by default. Via https://github.com/GoogleChrome/puppeteer/blob/master/lib/Launcher.js#L38

Functions

default_executable
Returns the path to Chrome’s executable.