pub struct Wait { /* private fields */ }Expand description
A helper to wait until some event has passed.
Implementations
Sourceimpl Wait
impl Wait
pub fn new(timeout: Duration, sleep: Duration) -> Self
pub fn with_timeout(timeout: Duration) -> Self
pub fn with_sleep(sleep: Duration) -> Self
pub fn forever() -> Self
Sourcepub fn until<F, G>(&self, predicate: F) -> Result<G, Timeout>
pub fn until<F, G>(&self, predicate: F) -> Result<G, Timeout>
Wait until the given predicate returns Some(G) or timeout arrives.
Note: If your predicate function shadows potential unexpected
errors you should consider using #strict_until.
Sourcepub fn strict_until<F, D, E, G>(&self, predicate: F, downcast: D) -> Result<G>
pub fn strict_until<F, D, E, G>(&self, predicate: F, downcast: D) -> Result<G>
Wait until the given predicate returns Ok(G), an unexpected error occurs or timeout arrives.
Errors produced by the predicate are downcasted by the additional provided closure.
If the downcast is successful - the error is ignored, otherwise the wait is terminated
and Err(error) containing the unexpected failure is returned to the caller.
You can use failure::Error::downcast::<YourStructName> out-of-the-box,
if you need to ignore one expected error, or you can implement a matching closure
that responds to multiple error types.